"""

This module contains tools for interacting with the MeerKAT archive

"""

import os, sys, glob, subprocess
from .startup import config
from . import jobs

#------------------------------------------------------------------------------------------------------------
def fetchFromArchive(captureBlockIdLink):
    """Fetch the dataset from the archive and write into the scratch area.

    Args:
        captureBlockIdLink (str): Link copied from https://archive.sarao.ac.za of the form
            https://archive-gw-1.kat.ac.za/captureBlockId/captureBlockId_sdp_l0.full.rdb?token=longTokenString.

    """

    # NOTE: We don't need routine any more
    captureBlockId=captureBlockIdLink.split("https://archive-gw-1.kat.ac.za/")[-1].split("/")[0]
    msPath=os.environ['BERK_MSCACHE']+os.path.sep+"%s_sdp_l0.ms" % (captureBlockId)
    cmd="mvftoms.py %s --flags cam,data_lost,ingest_rfi -o %s" % (captureBlockIdLink, msPath)
    os.system("screen -S fetch-%s -d -m %s" % (cmd))
    print("Fetching %s" % (msPath))

#------------------------------------------------------------------------------------------------------------
def checkFetchComplete(captureBlockId):
    """Check if the measurement set corresponding to the given captureBlockId has been fetched.

    Note:
        This relies on GNU Screen.

    """

    process=subprocess.run(['screen', '-ls'], universal_newlines = True,
                           stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
    if process.stdout.find("fetch-%s" % (captureBlockId)) == -1:
        return False
    else:
        return True

#------------------------------------------------------------------------------------------------------------
def stageMS(captureBlockId):
    """Sets up a measurement set for processing:

        1. Fetch from archive if necessary
        2. Unpack tar.gz if necessary and put measurement set into top-level of staging directory
        3. Return the path to the measurement set itself

    """

    # NOTE: MOVE THIS INTO BERK_UNPACK
    MSPath=config["stagingDir"]+os.path.sep+captureBlockId+"_sdp_l0.ms"
    if os.path.exists(MSPath) == False:
        pathToTGZ=fetchFromArchive(captureBlockId)
        topDir=os.getcwd()
        os.chdir(config['stagingDir'])
        #jobID=jobs.submitJob("tar -zxvf %s" % (pathToTGZ), "unpack")
        for p in os.walk("scratch"):
            if p[1][0] == os.path.split(MSPath)[-1]:
                break
        # NOTE: We'd need a job to do this
        mvCmd="mv %s/%s ." % (p[0], p[1][0])
        os.chdir(topDir)
    assert(os.path.exists(MSPath))







# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import sys, os
sys.path.insert(0, os.path.abspath('..'))
#sys.path.insert(0, os.path.abspath('../nemo'))
from berk import __version__

project = 'Berk'
copyright = '2023, Matt Hilton'
author = 'Matt Hilton'

# The short X.Y version
version = __version__
# The full version, including alpha/beta/rc tags
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'readthedocs_ext.readthedocs',
    'sphinxarg.ext'
]

source_suffix = '.rst'
language = 'en'
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



"""

This module contains tools for extracting info from MeerKAT images

"""

import os
import sys
import numpy as np
import astropy.io.fits as pyfits
import astropy.stats as apyStats
from astLib import *
from . import catalogs

#------------------------------------------------------------------------------------------------------------
def getImagesStats(imgFileName, radiusArcmin = 12):
    """Read the given MeerKAT image and return stats such as the image centre coords,
       effective frequency (GHz), RMS in uJy/beam, etc.

    Args:
        imgFileName (:obj:`str`): Path to the FITS images.
        radiusArcmin (:obj:`float`, optional): Radius in arcmin within which stats will
            be calculated.

    Returns:
        Dictionary of image statistics.

    """

    with pyfits.open(imgFileName) as img:
        d=img[0].data
        if d.ndim == 4:
            d=d[0, 0]
        assert(d.ndim == 2)
        wcs=astWCS.WCS(img[0].header, mode = 'pyfits')
    RADeg, decDeg=wcs.getCentreWCSCoords()
    RAMin, RAMax, decMin, decMax=astCoords.calcRADecSearchBox(RADeg, decDeg, radiusArcmin/60)
    clip=astImages.clipUsingRADecCoords(d, wcs, RAMin, RAMax, decMin, decMax)
    d=clip['data']
    wcs=clip['wcs']
    sigma=1e6
    for i in range(10):
        mask=np.logical_and(np.greater(d, d.mean()-3*sigma), np.less(d, d.mean()+3*sigma))
        sigma=np.std(d[mask])
    # print(">>> Image: %s - radiusArcmin = %.2f" % (sys.argv[1], radiusArcmin))
    # print("    clipped stdev image RMS = %.3f uJy/beam" % (sigma*1e6))
    # sbi=apyStats.biweight_scale(d, c = 9.0, modify_sample_size = True)
    # print("    biweight scale image RMS = %.3f uJy/beam" % (sbi*1e6))
    statsDict={'path': imgFileName,
               'object': wcs.header['OBJECT'],
               'centre_RADeg': RADeg,
               'centre_decDeg': decDeg,
               'RMS_uJy/beam': sigma*1e6,
               'freqGHz': wcs.header['CRVAL3']/1e9}

    return statsDict




"""

Startup routines and config for Berk

"""

import os, sys

on_rtd=os.environ.get('READTHEDOCS', None)
if on_rtd is not None:
    os.environ['BERK_ROOT']='.'
    os.environ['BERK_MSCACHE']='MSCache'
    os.environ['BERK_PLATFORM']='chpc'

# Settings are hard-coded for now, but could be put into a YAML config file later ---------------------------
config={}

if "BERK_ROOT" not in os.environ.keys():
    raise Exception("You need to set the BERK_ROOT environment variable - this defines the directory where all work will be done.")
if "BERK_MSCACHE" not in os.environ.keys():
    raise Exception("You need to set the BERK_MSCACHE environment variable - this defines the directory where retrieved measurement sets will be stored.")

os.makedirs(os.environ["BERK_MSCACHE"], exist_ok = True)
os.makedirs(os.environ["BERK_ROOT"], exist_ok = True)

# For using 'collect' to fetch data products from (potentially) multiple locations
if 'BERK_NODES_FILE' in os.environ.keys():
    config['nodesFile']=os.environ['BERK_NODES_FILE']
else:
    config['nodesFile']=None

# We can't just make this about the workload manager as it affects what we feed into oxkat
if 'BERK_PLATFORM' not in os.environ.keys():
    raise Exception("Set BERK_PLATFORM environment variable to either 'hippo' or 'chpc'")
if os.environ['BERK_PLATFORM'] == 'chpc':
    config['workloadManager']='pbs'
elif os.environ['BERK_PLATFORM'] == 'hippo':
    config['workloadManager']='slurm'
else:
    raise Exception("Environment variable BERK_PLATFORM is not set to 'hippo' or 'chpc'")

# Processing and data products will be written in sub-dirs here
config['rootDir']=os.environ["BERK_ROOT"]

# This should eventually be the scratch disk
# config['stagingDir']=config['rootDir']+os.path.sep+"staging"
# stagingDir="/scratch/month/mjh/"

# Directory where we do processing
config['processingDir']=config['rootDir']+os.path.sep+"processing"

# Directory where we archive data products
config['productsDir']=config['rootDir']+os.path.sep+"products"

# Directory where we'll cache e.g. oxkat
config['cacheDir']=config['rootDir']+os.path.sep+"cache"

# Oxkat version to use
# OLD: Tagged version
# config['oxkatVersion']="0.3"
# config['oxkatDir']=config['cacheDir']+os.path.sep+"oxkat-%s" % (config['oxkatVersion'])
# config['oxkatURL']="https://github.com/IanHeywood/oxkat/archive/refs/tags/v%s.tar.gz" % (config['oxkatVersion'])
# CURRENT: From Matt's git fork
config['oxkatVersion']="git"
config['oxkatDir']=config['cacheDir']+os.path.sep+"oxkat-%s" % (config['oxkatVersion'])
config['oxkatURL']="https://github.com/mattyowl/oxkat.git"

# Image-processing (source finding scripts from Jonah)
config['catalogScriptsDir']=config['cacheDir']+os.path.sep+"catalog-scripts"

print("Using oxkat version: %s" % (config['oxkatVersion']))
if config['oxkatVersion'] == "git":
    print("Remember to remove %s before running berk if you need to fetch an updated version from the git repository" % (config['oxkatDir']))

# Set-up ----------------------------------------------------------------------------------------------------
if on_rtd is None:
    dirsToMake=[config['processingDir'], config['productsDir'], config['cacheDir']]
    for d in dirsToMake:
        os.makedirs(d, exist_ok = True)

    if os.path.exists(config['oxkatDir']) == False:
        topDir=os.getcwd()
        os.chdir(config['cacheDir'])
        if config['oxkatVersion'] != 'git':
            os.system("wget %s" % (config['oxkatURL']))
            os.system("tar -zxvf v%s.tar.gz" % (config['oxkatVersion']))
        else:
            os.system("git clone %s oxkat-%s" % (config['oxkatURL'], config['oxkatVersion']))
        os.chdir(topDir)

    if os.path.exists(config['catalogScriptsDir']) == False:
        os.system("git clone https://github.com/mattyowl/Image-processing %s" % (config["catalogScriptsDir"]))




"""

Tools for running jobs via a workload manager (Slurm or PBS).

"""

import subprocess, os, sys

SLURM_TEMPLATE="""#!/bin/sh
#SBATCH --nodes=$NODES
#SBATCH --ntasks=$TASKS
#SBATCH --cpus-per-task=$CPUSPERTASK
#SBATCH --mem=$MEM
#SBATCH --time=$TIME
#SBATCH --output=$JOBNAME.log
#SBATCH --error=$JOBNAME.err

$CMD
sleep 10
"""

PBS_TEMPLATE="""#!/bin/sh
#PBS -l nodes=$NODES:ppn=$TASKS,mem=$MEM
#PBS -P $PBS_PROJECT
#PBS -q $PBS_QUEUE
#PBS -l walltime=$TIME
#PBS -o $JOBNAME.log
#PBS -e $JOBNAME.err
#PBS -m abe
#PBS -M $PBS_EMAIL
ulimit -s unlimited
cd $CWD

$CMD
sleep 10
"""

def writeJobScript(cmd, jobName, nodes = 1, tasks = 1, mem = 8000, time = "12:00:00",
                   workloadManager = 'slurm'):
    """Write a job script to run a command via slurm.

    Args:
        cmd (str): The command to run.
        jobName (str): Name of the job (used for e.g. log files).
        nodes (int): Number of nodes on which the job will run.
        tasks (int): Number of tasks to run per node.
        mem (int): Requested memory (KB) for the job.
        time (str): Wall time limit for the job.
        workloadManager (str): Either 'slurm' or 'pbs'. Note that for PBS, several
            environment variables also need to be defined (`PBS_NODETYPE`,
            `PBS_PROJECT`, `PBS_QUEUE`, `PBS_EMAIL`).

    Returns:
        The file name for the batch file.

    Note:
        This doesn't submit jobs... see ``submitJob`` for a routine that does (it uses this routine).

    """

    if workloadManager == 'slurm':
        script=SLURM_TEMPLATE
    elif workloadManager == 'pbs':
        script=PBS_TEMPLATE
    else:
        raise Exception("workloadManager should be either 'slurm' or 'pbs'")
    script=script.replace("$CMD", cmd)
    script=script.replace("$JOBNAME", jobName)
    script=script.replace("$NODES", str(nodes))
    script=script.replace("$TASKS", str(tasks))
    if workloadManager == 'slurm':
        script=script.replace("$MEM", str(mem)) # MB
    elif workloadManager == 'pbs':
        script=script.replace("$MEM", "%.dgb" % (mem/1000))
    script=script.replace("$TIME", str(time))
    if workloadManager == 'pbs':
        envVars=['PBS_NODETYPE', 'PBS_PROJECT', 'PBS_QUEUE', 'PBS_EMAIL']
        for e in envVars:
            script=script.replace("$%s" % (e), os.environ[e])
        script=script.replace("$CWD", os.path.abspath(os.path.curdir))

    fileName=jobName+"."+workloadManager
    with open(fileName, "w") as outFile:
        outFile.write(script)

    return fileName


def submitJob(cmd, jobName, dependentJobIDs = None, nodes = 1, tasks = 1, mem = 8000,
              time = "12:00:00", workloadManager = 'slurm', cmdIsBatchScript = False):
    """Submit a command to run on a node via Slurm or PBS.

    Args:
        cmd (str): The command to run.
        jobName (str): Name of the job (used for e.g. log files).
        dependentJobIDs (list): If this job depends on a previous job completing sucessfully, give the
            job ID numbers here, as a list.
        nodes (int): Number of nodes on which the job will run.
        tasks (int): Number of tasks to run per node.
        mem (int): Requested memory (KB) for the job.
        time (str): Wall time limit for the job.
        workloadManager (str): Either 'slurm' or 'pbs'. Note that for PBS, several
            environment variables also need to be defined (`PBS_NODETYPE`,
            `PBS_PROJECT`, `PBS_QUEUE`, `PBS_EMAIL`).

    Returns:
        The ID number of the submitted Slurm job.

    Note:
        The job is submitted from the current working directory. So you *might*
        want to use absolute paths here.

    """

    if cmdIsBatchScript == True:
        fileName=cmd
    else:
        fileName=writeJobScript(cmd, jobName, nodes = nodes, tasks = tasks, mem = mem, time = time,
                                workloadManager = workloadManager)

    if workloadManager == 'slurm':
        args=['sbatch']
        if dependentJobIDs is not None:
            if type(dependentJobIDs) != list:
                raise Exception("dependentJobIDs must be given as a list")
            dependStr="afterok"
            for dependentJobID in dependentJobIDs:
                dependStr=dependStr+":"+str(dependentJobID)
            args=args+['--dependency=%s' % (dependStr)]
        args=args+[fileName]
    elif workloadManager == 'pbs':
        args=['qsub']
        if dependentJobIDs is not None:
            if type(dependentJobIDs) != list:
                raise Exception("dependentJobIDs must be given as a list")
            dependStr="afterok"
            for dependentJobID in dependentJobIDs:
                dependStr=dependStr+":"+str(dependentJobID)
            args=args+['-W depend=%s' % (dependStr)]
        args=args+[fileName]
    else:
        raise Exception("workloadManager should be either 'slurm' or 'pbs'")

    process=subprocess.run(args, universal_newlines=True, stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT)
    if process.returncode != 0:
        raise Exception("Non-zero return code when submitting job %s" % (jobName))
    if workloadManager == 'slurm':
        assert(process.stdout[:19] == "Submitted batch job")
        jobID=int(process.stdout.split("Submitted batch job")[-1])
    elif workloadManager == 'pbs':
        jobID=int(process.stdout.split(".")[0])

    return jobID




"""

Berk - a package for producing and managing MeerKAT continuum survey data.

"""

#from . import catalogs 
from . import startup

#__all__ = ['startUp']

from . import _version
__version__ = _version.get_versions()['version']



"""

This module contains tools for handling catalogs, which are usually :obj:`astropy.table.Table` objects.

"""

from astLib import *
import numpy as np
import operator
import os
import sys
import time
import astropy.table as atpy
from astropy.coordinates import SkyCoord
from astropy.coordinates import match_coordinates_sky
import astropy.io.fits as pyfits
from scipy import ndimage
from . import __version__

# For adding meta data to output
import datetime

#------------------------------------------------------------------------------------------------------------
def catalog2DS9(catalog, outFileName, constraintsList = [], addInfo = [], idKeyToUse = 'name',\
                RAKeyToUse = 'RADeg', decKeyToUse = 'decDeg', color = "cyan", showNames = True,\
                writeBerkInfo = True, coordSys = 'fk5', regionShape = 'point', width = 1):
    """Writes a DS9 region file corresponding to the given catalog. 
    
    Args:
        catalog (:obj:`astropy.table.Table`): An astropy Table where each row represents an object.
        outFileName (:obj:`str`): A file name for the output DS9 region file.
        constraintsList (:obj:`list`, optional): A list of constraints in the same format as used by 
            :func:`selectFromCatalog`.
        addInfo (:obj:`list`, optional): A list of dictionaries with keys named `key` and `fmt` (e.g., 
            ``{'key': "SNR", 'fmt': "%.3f"}``). These will be added to the object label shown in DS9.
        idKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that defines the 
            object's name. Used to label objects in the DS9 region file.
        RAKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that contains the 
            RA of the object in decimal degrees.
        decKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that contains the
            declination of the object in decimal degrees.
        color (:obj:`str`, optional): The color of the plot symbol used by DS9.
        writeBerkInfo (:obj:`bool`, optional): If ``True``, writes a line with the `berk` version and date
            generated at the top of the DS9 .reg file.
        coordSys (:obj:`str`, optional): A string defining the coordinate system used for RA, dec, as 
            understood by DS9.
        
    Returns:
        None
    
    """
    
    cutCatalog=selectFromCatalog(catalog, constraintsList) 
    
    with open(outFileName, "w") as outFile:
        timeStamp=datetime.datetime.today().date().isoformat()
        comment="# DS9 region file"
        if writeBerkInfo == True:
            comment=comment+" generated by Berk (version: %s on %s)\n" % (__version__, timeStamp)
        else:
            comment=comment+"\n"
        outFile.write(comment)
        outFile.write('global dashlist=8 3 width=%d font="helvetica 10 normal" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n' % (width))
        for obj in cutCatalog:
            if len(addInfo) > 0:
                infoString=""
                for d in addInfo:
                    if infoString != "":
                        infoString=infoString+" "
                    if obj[d['key']] != None:
                        infoString=infoString+d['fmt'] % (obj[d['key']])
                    else:
                        infoString=infoString+"%s" % (str(obj[d['key']]))
                infoString=" ["+infoString+"]"
            else:
                infoString=""
            if color == 'key':
                colorString=obj['color']
            else:
                colorString=color
            if showNames == True:
                infoString=str(obj[idKeyToUse])+infoString
            if regionShape == 'point':
                outFile.write("%s;point(%.6f,%.6f) # point=cross color={%s} text={%s}\n" \
                            % (coordSys, obj[RAKeyToUse], obj[decKeyToUse], colorString, infoString))
            elif regionShape == 'circle':
                outFile.write('%s;circle(%.6f,%.6f,360") # color={%s} text={%s}\n' \
                            % (coordSys, obj[RAKeyToUse], obj[decKeyToUse], colorString, infoString))                

#------------------------------------------------------------------------------------------------------------
def makeName(RADeg, decDeg, prefix = 'MKT'):
    """Makes an object name string from the given object coordinates, following the IAU convention.
    
    Args:
        RADeg (:obj:`float`): Right ascension of the object in J2000 decimal degrees.
        decDeg (:obj:`float`): Declination of the object in J2000 decimal degrees.
        prefix (:obj:`str`, optional): Prefix for the object name.
    
    Returns:
        Object name string in the format `prefix JHHMM.m+/-DDMM`.
    
    """
    
    actName=prefix+" J"+_makeRA(RADeg)+_makeDec(decDeg)
    
    return actName

#------------------------------------------------------------------------------------------------------------
def makeLongName(RADeg, decDeg, prefix = "MKT"):
    """Makes a long format object name string from the given object coordinates, following the IAU convention.
    
    Args:
        RADeg (:obj:`float`): Right ascension of the object in J2000 decimal degrees.
        decDeg (:obj:`float`): Declination of the object in J2000 decimal degrees.
        prefix (:obj:`str`, optional): Prefix for the object name.
    
    Returns:
        Object name string in the format `prefix JHHMMSS.s+/-DDMMSS`.
    
    """
    
    actName=prefix+" J"+_makeLongRA(RADeg)+_makeLongDec(decDeg)
    
    return actName
    
#------------------------------------------------------------------------------------------------------------
def _makeRA(myRADeg):
    """Makes RA part of ACT names.
    
    """
    hours=(myRADeg/360)*24
    strHours=("%.10f" % (hours))
    if hours<10:
        sHours="0"+strHours[0]
    else:
        sHours=strHours[:2]
    
    mins=float(strHours[strHours.index("."):])*60
    strMins=("%.10f" % (mins))
    if mins < 10:
        sMins="0"+strMins[:3]
    else:
        sMins=strMins[:4]

    return (sHours+sMins)#[:-2] # Trims off .x as not used in ACT names
        
#------------------------------------------------------------------------------------------------------------
def _makeDec(myDecDeg):
    """Makes dec part of ACT names
    
    """
    
    # Positive
    if myDecDeg>0:
        if myDecDeg<10:
            sDeg="0"+str(myDecDeg)[0]
        else:
            sDeg=str(myDecDeg)[:2]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
        
        return "+"+sDeg+sMins
    else:
        if myDecDeg>-10:
            sDeg="-0"+str(myDecDeg)[1]
        else:
            sDeg=str(myDecDeg)[:3]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
        
        return str(sDeg+sMins)

#-------------------------------------------------------------------------------------------------------------
def _makeLongRA(myRADeg):
    """Make a long RA string, i.e. in style of long XCS names
    
    """
    
    hours=(myRADeg/360)*24
    if hours<10:
        sHours="0"+str(hours)[0]
    else:
        sHours=str(hours)[:2]
    
    mins=float(str(hours)[str(hours).index("."):])*60
    if mins<10:
        sMins="0"+str(mins)[0]
    else:
        sMins=str(mins)[:2]
        
    secs=float(str(mins)[str(mins).index("."):])*60
    if secs<10:
        sSecs="0"+str(secs)[:3]
    else:
        sSecs=str(secs)[:4]     
        
    return sHours+sMins+sSecs
        
#-------------------------------------------------------------------------------------------------------------
def _makeLongDec(myDecDeg):
    """Make a long dec sting i.e. in style of long XCS names
    
    """
    # Positive
    if myDecDeg>0:
        if myDecDeg<10:
            sDeg="0"+str(myDecDeg)[0]
        else:
            sDeg=str(myDecDeg)[:2]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
            
        secs=float(str(mins)[str(mins).index("."):])*60
        if secs<10:
            sSecs="0"+str(secs)[:3]
        else:
            sSecs=str(secs)[:4]         
        
        return "+"+sDeg+sMins+sSecs
    else:
        if myDecDeg>-10:
            sDeg="-0"+str(myDecDeg)[1]
        else:
            sDeg=str(myDecDeg)[:3]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
            
        secs=float(str(mins)[str(mins).index("."):])*60
        if secs<10:
            sSecs="0"+str(secs)[:3]
        else:
            sSecs=str(secs)[:4]         
        
        return sDeg+sMins+sSecs
        
#-------------------------------------------------------------------------------------------------------------
def selectFromCatalog(catalog, constraintsList):
    """Return a table of objects matching the given constraints from the catalog. 
    
    Args:
        catalog (:obj:`astropy.table.Table`): The catalog from which objects will be selected.
        constraintsList (:obj:`list`): A list of constraints, where each item is a string of the form
            "key < value", "key > value", etc.. Note that the spaces between the key, operator 
            (e.g. '<'), and value are essential.
    
    Returns:
        An astropy Table object.
    
    """
            
    passedConstraint=catalog
    for constraintString in constraintsList:
        key, op, value=constraintString.split()
        passedConstraint=passedConstraint[eval("passedConstraint['%s'] %s %s" % (key, op, value))]

    return passedConstraint

#------------------------------------------------------------------------------------------------------------
def removeDuplicates(tab):
    """Removes duplicate objects from the catalog - keeping the highest SNR detection for each duplicate. 
    This routine is used to clean up the output of MPI runs (where we have overlapping tiles).
    
    Args:
        tab (:obj:`astropy.table.Table`): The object catalog to be checked for duplicates.

    Returns:
        Table with duplicates removed (:obj:`astropy.table.Table`), the number of duplicates found, and a
        list of names for the duplicated objects.
    
    """

    if len(tab) == 1:
        return tab, 1, []
    
    # Find all duplicates
    cat=SkyCoord(ra = tab['RADeg'].data, dec = tab['decDeg'].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat, cat, nthneighbor = 2)
    mask=np.less(rDeg.value, XMATCH_RADIUS_DEG)
    noDupMask=np.greater_equal(rDeg.value, XMATCH_RADIUS_DEG)
    dupTab=tab[mask]
    noDupTab=tab[noDupMask]
    
    # All duplicates removed?
    if mask.sum() == 0:
        return tab, 0, []
    
    # Much faster
    keepMask=np.zeros(len(dupTab), dtype = bool)
    for i in range(len(dupTab)):
        # NOTE: astCoords does not like atpy.Columns sometimes...
        rDeg=astCoords.calcAngSepDeg(dupTab['RADeg'][i], dupTab['decDeg'][i], dupTab['RADeg'].data, dupTab['decDeg'].data)
        mask=np.less_equal(rDeg, XMATCH_RADIUS_DEG)
        if mask.sum() == 0:	# This ought not to be possible but catch anyway
            bestIndex=i
        else:
            indices=np.where(mask == True)[0]
            bestIndex=indices[np.equal(dupTab['SNR'][mask], dupTab['SNR'][mask].max())][0]
        keepMask[bestIndex]=True
    keepTab=dupTab[keepMask]
    
    keepTab=atpy.vstack([keepTab, noDupTab])
    keepTab.sort('RADeg')
    
    return keepTab, len(dupTab), dupTab['name']

#------------------------------------------------------------------------------------------------------------
def crossMatch(refCatalog, matchCatalog, radiusArcmin = 2.5):
    """Cross matches `matchCatalog` onto `refCatalog` for objects found within some angular radius 
    (specified in arcmin).
    
    Args:
        refCatalog (:obj:`astropy.table.Table`): The reference catalog.
        matchCatalog (:obj:`astropy.table.Table`): The catalog to match onto the reference catalog.
        radiusArcmin (:obj:`float`, optional): Cross-match radius in arcmin.
    
    Returns:
        Cross-matched reference catalog, matchCatalog, and array of angular separation in degrees, for 
        objects in common within the matching radius. The cross matched columns are sorted such that rows in
        each correspond to the matched objects.
    
    """
    
    inTab=refCatalog
    outTab=matchCatalog
    RAKey1, decKey1=getTableRADecKeys(inTab)
    RAKey2, decKey2=getTableRADecKeys(outTab)
    cat1=SkyCoord(ra = inTab[RAKey1].data, dec = inTab[decKey1].data, unit = 'deg')
    xMatchRadiusDeg=radiusArcmin/60.
    cat2=SkyCoord(ra = outTab[RAKey2].data, dec = outTab[decKey2].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat1, cat2, nthneighbor = 1)
    mask=np.less(rDeg.value, xMatchRadiusDeg)  
    matched_outTab=outTab[xIndices]
    inTab=inTab[mask]
    matched_outTab=matched_outTab[mask]
    rDeg=rDeg.value[mask]
    
    return inTab, matched_outTab, rDeg

#------------------------------------------------------------------------------------------------------------
def removeCrossMatched(refCatalog, matchCatalog, radiusArcmin = 2.5):
    """Cross matches `matchCatalog` onto `refCatalog` for objects found within some angular radius 
    (specified in arcmin), and returns `refCatalog` with the matching entries removed.
    
    Args:
        refCatalog (:obj:`astropy.table.Table`): The reference catalog.
        matchCatalog (:obj:`astropy.table.Table`): The catalog to match onto the reference catalog.
        radiusArcmin (:obj:`float`, optional): Cross-match radius in arcmin.
    
    Returns:
        Cross-matched reference catalog (:obj:`astropy.table.Table`) with matches to `matchCatalog` removed.
        
    """
        
    inTab=refCatalog
    outTab=matchCatalog
    RAKey1, decKey1=getTableRADecKeys(inTab)
    RAKey2, decKey2=getTableRADecKeys(outTab)
    cat1=SkyCoord(ra = inTab[RAKey1].data, dec = inTab[decKey1].data, unit = 'deg')
    xMatchRadiusDeg=radiusArcmin/60.
    cat2=SkyCoord(ra = outTab[RAKey2].data, dec = outTab[decKey2].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat1, cat2, nthneighbor = 1)
    mask=np.greater(rDeg.value, xMatchRadiusDeg)  
    inTab=inTab[mask]
    
    return inTab
    
#------------------------------------------------------------------------------------------------------------
def getTableRADecKeys(tab):
    """Returns the column names in the table in which RA, dec coords are stored, after trying a few possible
    name variations.
    
    Args:
        tab (:obj:`astropy.table.Table`): The table to search.
        
    Returns:
        Name of the RA column, name of the dec. column
    
    """
    RAKeysToTry=['ra', 'RA', 'RADeg']
    decKeysToTry=['dec', 'DEC', 'decDeg', 'Dec']
    RAKey, decKey=None, None
    for key in RAKeysToTry:
        if key in tab.keys():
            RAKey=key
            break
    for key in decKeysToTry:
        if key in tab.keys():
            decKey=key
            break
    if RAKey is None or decKey is None:
        raise Exception("Couldn't identify RA, dec columns in the supplied table.")
    
    return RAKey, decKey

#------------------------------------------------------------------------------------------------------------
def getCatalogWithinImage(tab, shape, wcs, mask = None):
    """Returns the subset of the catalog with coordinates within the image defined by the given `shape`, 
    `wcs`. Optionally, a `mask` may also be applied.
    
    Args:
        tab (:obj:`astropy.table.Table`): Catalog, as an astropy Table object. Must have columns called 
            'RADeg', 'decDeg' that contain object coordinates in decimal degrees.
        shape (:obj:`list`): Shape of the array corresponding to the image / map.
        wcs (:obj:`astWCS.WCS`): WCS of the image.
        mask (optional, :obj:`np.ndarray`): Mask with same dimensions and WCS as the image. Pixels with
            value = 1 indicate valid area, and pixels with value = 0 are considered to be outside the mask.
            If this is given, the returned catalog will contain only objects in the valid area defined by
            this image mask.
    
    Returns:
        An astropy Table containing the subset of objects within the image.
    
    """
    
    xyCoords=np.array(wcs.wcs2pix(tab['RADeg'].tolist(), tab['decDeg'].tolist()))
    selected=[]
    for i in range(len(tab)):
        x, y=xyCoords[i][0], xyCoords[i][1]
        if np.isnan(x) == True or np.isnan(y) == True:
            selected.append(False)
            continue
        if x >= 0 and x < shape[1]-1 and y >= 0 and y < shape[0]-1:
            if mask is not None:
                if mask[int(round(y)), int(round(x))] == 1:
                    selected.append(True)
                else:
                    selected.append(False)
            else:
                selected.append(True)
        else:
            selected.append(False)
    
    return tab[selected]

#------------------------------------------------------------------------------------------------------------
def addFootprintColumnToCatalog(tab, label, areaMask, wcs):
    """Add `footprint_label` column to the catalog, flagging objects found within the valid area of the given
    mask.

    Args:
        tab (:obj:`astropy.table.Table`): Catalog, as an astropy Table object. Must have columns called
            'RADeg', 'decDeg' that contain object coordinates in decimal degrees.
        label (:obj:`str`): A column named `footprint_label` will be added to the catalog. Objects in the
            catalog that fall within the valid area of the given area mask will have `footprint_label`
            set to True.
        areaMask (:obj:`np.ndarray`): Mask image defining the footprint corresponding to the given WCS.
            Pixels with value = 1 indicate valid area, and pixels with value = 0 are considered to be
            outside the mask.
        wcs (:obj:`astWCS.WCS`): WCS of the area mask that defines the footprint.

    Returns:
        An astropy Table with `footprint_label` column added.

    """

    inMask=np.zeros(len(tab['RADeg'].data), dtype = bool)
    coords=wcs.wcs2pix(tab['RADeg'].data, tab['decDeg'].data)
    coords=np.array(np.round(coords), dtype = int)
    mask1=np.logical_and(coords[:, 0] >= 0, coords[:, 1] >= 0)
    mask2=np.logical_and(coords[:, 0] < areaMask.shape[1], coords[:, 1] < areaMask.shape[0])
    mask=np.logical_and(mask1, mask2)
    inMask[mask]=inMask[mask]+areaMask[coords[:, 1][mask], coords[:, 0][mask]]
    tab['footprint_%s' % (label)]=inMask

    return tab



"""

Tasks that can be performed by Berk

"""

import os
import sys
import subprocess
import glob
import time
import datetime
import astropy.table as atpy
from . import startup, archive, jobs, catalogs, images,  __version__

#------------------------------------------------------------------------------------------------------------
def fetch(captureBlockId):
    """Fetch...

    """

    captureBlockIdLink=captureBlockId
    captureBlockId=captureBlockIdLink.split("https://archive-gw-1.kat.ac.za/")[-1].split("/")[0]
    msPath=os.environ['BERK_MSCACHE']+os.path.sep+"%s_sdp_l0.ms" % (captureBlockId)
    if archive.checkFetchComplete(captureBlockId) == False:
        cmd="mvftoms.py %s --flags cam,data_lost,ingest_rfi -o %s" % (captureBlockIdLink, msPath)
        os.system(cmd)
        # print("Run this command in GNU screen:")
        # print(cmd)
        # print("[yes, this is clunky - but automatically running in screen isn't working at the moment]")
        # os.system("screen -dmS fetch-%s bash -c '%s'" % (captureBlockId, cmd))
        # print("Fetching %s" % (msPath))
    else:
        print("Already fetching %s" % (msPath))
    sys.exit()

#------------------------------------------------------------------------------------------------------------
def store(captureBlockId):
    """Store...

    """

    print("Task 'store' is not implemented yet.")
    sys.exit()

#------------------------------------------------------------------------------------------------------------
def _getBandKey(freqGHz):
    """Return band name ('L' or 'UHF') based on freqGHz.

    """
    if freqGHz > 1.2 and freqGHz < 1.3:
        bandKey='L'
    else:
        print("implement UHF cut")
        import IPython
        IPython.embed()
        sys.exit()

    return bandKey

#------------------------------------------------------------------------------------------------------------
def builddb():
    """Build database...

    """

    # Build global catalog in each band (L-band, UHF)
    globalTabsDict={'L': None, 'UHF': None}
    tabFilesList=glob.glob(startup.config['productsDir']+os.path.sep+'catalogs'+os.path.sep+'*_bdsfcat.fits')
    for t in tabFilesList:
        if t.find("srl_bdsfcat") == -1:
            tab=atpy.Table().read(t)
            freqGHz=tab.meta['FREQ0']/1e9
            bandKey=_getBandKey(freqGHz)
            if globalTabsDict[bandKey] is None:
                globalTabsDict[bandKey]=tab
            else:
                globalTabsDict[bandKey]=atpy.vstack([globalTabsDict[bandKey], tab])

    # To implement: remove duplicates [we may want to make time series?]

    # Sorting, additional meta data, output
    for bandKey in globalTabsDict.keys():
        if globalTabsDict[bandKey] is not None:
            outFileName=startup.config['productsDir']+os.path.sep+"survey_catalog_%s.fits" % (bandKey)
            globalTabsDict[bandKey].sort('DEC')
            globalTabsDict[bandKey].sort('RA')
            globalTabsDict[bandKey].meta['BAND']=bandKey
            globalTabsDict[bandKey].meta['BERKVER']=__version__
            globalTabsDict[bandKey].meta['DATEMADE']=datetime.date.today().isoformat()
            globalTabsDict[bandKey].write(outFileName, overwrite = True)
            catalogs.catalog2DS9(globalTabsDict[bandKey], outFileName.replace(".fits", ".reg"),
                                 idKeyToUse = 'Source_name', RAKeyToUse = 'RA', decKeyToUse = 'DEC')
            print("Wrote %s" % (outFileName))

    # Make image table - centre coords, radius [approx.], RMS, band, image path - UHF and L together.
    # Report command (when we make it) could load and dump some of that info
    outFileName=startup.config['productsDir']+os.path.sep+"images.fits"
    imgFilesList=glob.glob(startup.config['productsDir']+os.path.sep+"images"+os.path.sep+"pbcorr_*.fits")
    statsDictList=[]
    for imgFile in imgFilesList:
        statDict=images.getImagesStats(imgFile)
        statDict['path']=statDict['path'].replace(startup.config['productsDir']+os.path.sep, '')
        statDict['band']=_getBandKey(statDict['freqGHz'])
        statsDictList.append(statDict)
    imgTab=atpy.Table()
    for key in statDict.keys():
        arr=[]
        for s in statsDictList:
            arr.append(s[key])
        imgTab[key]=arr
    imgTab.meta['BERKVER']=__version__
    imgTab.meta['DATEMADE']=datetime.date.today().isoformat()
    imgTab.write(outFileName, overwrite = True)
    print("Wrote %s" % (outFileName))

    # Generate survey mask in some format - we'll use that to get total survey area


#------------------------------------------------------------------------------------------------------------
def collect():
    """Collect...

    """

    print("Collecting processed data products...")
    if 'BERK_NODES_FILE' not in os.environ.keys():
        print("You need to set the BERK_NODES_FILE environment variable to use the 'collect' task.")
        sys.exit()
    nodesFilePath=os.environ['BERK_NODES_FILE']
    try:
        stubs=[]
        with open(os.environ['BERK_NODES_FILE'], "r") as inFile:
            for line in inFile.readlines():
                stubs.append(line)
    except:
        import urllib.request  # the lib that handles the url stuff
        stubs=[]
        for line in urllib.request.urlopen(os.environ['BERK_NODES_FILE']):
            l=line.decode('utf-8')
            if l[0] != "#" and len(l) > 3:
                stubs.append(l.strip())

    # Get images
    print("Collecting images...")
    toPath=startup.config['productsDir']+os.path.sep+"images"
    os.makedirs(toPath, exist_ok = True)
    for s in stubs:
        imgPath="processing/*/IMAGES/pbcorr*pcalmask-MFS-image.fits"
        cmd="rsync -avP %s%s %s" % (s, os.path.sep+imgPath, toPath)
        os.system(cmd)

    # Get catalogs
    print("Collecting catalogs...")
    toPath=startup.config['productsDir']+os.path.sep+"catalogs"
    os.makedirs(toPath, exist_ok = True)
    for s in stubs:
        catPath="processing/*/IMAGES/pbcorr_trim_*_pybdsf/*_bdsfcat.fits"
        cmd="rsync -avP %s%s %s" % (s, os.path.sep+catPath, toPath)
        os.system(cmd)

    print("Finished!")
    sys.exit()

#------------------------------------------------------------------------------------------------------------
def process(captureBlockId):
    """Process...

    """

    # Forget staging dir, just do a symbolic link to the MSCache dir
    MSPath=os.environ['BERK_MSCACHE']+os.path.sep+captureBlockId+"_sdp_l0.ms"

    # Setup in processing dir
    MSProcessDir=startup.config['processingDir']+os.path.sep+captureBlockId
    if os.path.exists(MSProcessDir) == True:
        raise Exception("Processing directory %s exists and is not empty - remove it and re-run, if you're sure you don't need its contents." % (MSProcessDir))
    os.makedirs(MSProcessDir)
    os.chdir(MSProcessDir)
    os.system("ln -s %s" % (os.path.abspath(MSPath)))
    oxdirs=['setups', 'tools', 'oxkat', 'data']
    for oxdir in oxdirs:
        os.system("ln -s %s" % (startup.config['oxkatDir']+os.path.sep+oxdir))

    # Generate the oxkat job scripts then spin through + submit them ourselves
    # 1GC
    os.system("python3 setups/1GC.py %s" % os.environ['BERK_PLATFORM'])
    jobCmds=[]
    dependent=[]
    with open("submit_1GC_jobs.sh") as inFile:
        lines=inFile.readlines()
        for line in lines:
            if line.find("sbatch") != -1 and startup.config['workloadManager'] == 'slurm':
                sbatchCmd=line[line.find("sbatch") :].split(" |")[0]
                if sbatchCmd.find("-d afterok:") != -1:
                    sbatchCmd=sbatchCmd.split("}")[-1].strip()
                    dependent.append(True)
                else:
                    sbatchCmd=sbatchCmd.split("sbatch")[-1].strip()
                    dependent.append(False)
                jobCmds.append(sbatchCmd)
            elif line.find("qsub") != -1 and startup.config['workloadManager'] == 'pbs':
                qsubCmd=line[line.find("qsub") :].split(" |")[0]
                if qsubCmd.find("-W depend=afterok") != -1:
                    qsubCmd=qsubCmd.split("}")[-1].strip()
                    dependent.append(True)
                else:
                    qsubCmd=qsubCmd.split("qsub")[-1].strip()
                    dependent.append(False)
                jobCmds.append(qsubCmd)

    jobIDs=[]
    for cmd, dep in zip(jobCmds, dependent):
        if dep == False:
            dependentJobIDs=None
        else:
            dependentJobIDs=jobIDs
        jobName=os.path.split(cmd)[-1]
        jobID=jobs.submitJob(cmd, jobName, dependentJobIDs = dependentJobIDs, workloadManager = startup.config['workloadManager'], cmdIsBatchScript = True)
        jobIDs.append(jobID)

    # Run the FLAG and 2GC setup scripts as a job, then chain them together
    cmd="python3 setups/FLAG.py %s" % (os.environ['BERK_PLATFORM'])
    jobID=jobs.submitJob(cmd, "SETUP_FLAG_JOBS", dependentJobIDs = jobIDs, workloadManager = startup.config['workloadManager'])
    jobIDs.append(jobID)
    cmd="python3 setups/2GC.py %s" % (os.environ['BERK_PLATFORM'])
    jobID=jobs.submitJob(cmd, "SETUP_2GC_JOBS", dependentJobIDs = jobIDs, workloadManager = startup.config['workloadManager'])
    jobIDs.append(jobID)
    cmd="berk_chain %s submit_flag_jobs.sh submit_2GC_jobs.sh" % (startup.config['workloadManager'])
    jobID=jobs.submitJob(cmd, "CHAIN_FLAG+2GC_JOBS", dependentJobIDs = jobIDs, workloadManager = startup.config['workloadManager'])
    print("All jobs submitted")
    sys.exit()

#------------------------------------------------------------------------------------------------------------
def analyse(captureBlockId):
    """Analyse...

    """

    # Setup in processing dir
    MSProcessDir=startup.config['processingDir']+os.path.sep+captureBlockId
    if os.path.exists(MSProcessDir) == False:
        raise Exception("Processing directory %s does not exist - you need to process the data before the analyse task will run." % (MSProcessDir))
    os.chdir(MSProcessDir)
    os.system("ln -s %s" % (startup.config["catalogScriptsDir"]+os.path.sep+"sourcefinding.py"))
    os.system("ln -s %s" % (startup.config["catalogScriptsDir"]+os.path.sep+"catalog_matching.py"))
    os.system("ln -s %s" % (startup.config["catalogScriptsDir"]+os.path.sep+"parsets"))

    # Source finding is fairly lightweight so we put everything in one job script
    # We will have issues with needing to see the internet to fetch cross match catalogs though
    # So we will need to cache NVSS catalogs for a given direction when doing 'stage'
    # OR forget using catalog_matching.py here and just do later with database/catalog scripts
    imgPaths=glob.glob("IMAGES/*pcalmask-MFS-image.fits")
    for i in imgPaths:
        if i.find("pbcorr_trim") != -1:
            continue
        imgPath=os.path.abspath(i)
        cmd="mkat_primary_beam_correct %s -T" % (imgPath)

        imgDir, imgFileName=os.path.split(imgPath)
        pbcorrImgPath=imgDir+os.path.sep+"pbcorr_trim_"+imgFileName
        cmd=cmd+"\npython3 sourcefinding.py c %s -o fits --survey MSS" % (pbcorrImgPath)

        # The bit below isn't going to work on compute nodes - so we may as well move this
        label=imgFileName.split(".ms_")[0].split(".")[0]
        catPath=imgDir+os.path.sep+"pbcorr_trim_"+label+"_pybdsf"+os.path.sep+"pbcorr_trim_"+label+"_bdsfcat.fits"
        cmd=cmd+"\npython3 catalog_matching.py %s NVSS --astro --flux" % (catPath)

        jobID=jobs.submitJob(cmd, 'source-finding-%s' % (imgFileName), dependentJobIDs = None, nodes = 1, tasks = 20, mem = 64000,
                                time = "24:00:00", cmdIsBatchScript = False, workloadManager = startup.config['workloadManager'])
        print("Submitted source finding and analysis job %d" % (jobID))
    sys.exit()






import os
import glob
from setuptools import setup
from setuptools import Extension
import versioneer

setup(name='berk',
      version=versioneer.get_version(),
      author='Matt Hilton',
      author_email='matt.hilton@wits.ac.za',
      packages=['berk'],
      scripts=['bin/berk', 'bin/berk_chain', 'bin/mkat_primary_beam_correct'],
)
